
import pytest

from .abstract import (
    CardABC,
    CardPropertyABC,
)


def test_empty_cards():
    
    class SimpleCard(CardABC):
        pass

    SimpleCard()


def test_card_definition_with_one_property():

    class TestProperty(CardPropertyABC):
        AcceptableValues = {1, 2, 3, 4}

    class TestCard(CardABC):
        CardProperties = [TestProperty('value')]

    with pytest.raises(ValueError):
        TestCard(test_value='break')

    with pytest.raises(ValueError):
        TestCard(value=5)

    TestCard(value=1)


@staticmethod  # must be a static method to prevent becoming bound
def is_string(s):
    return isinstance(s, str)


def test_card_definition_with_many_properties():

    class NameProperty(CardPropertyABC):
        AcceptableValues = is_string

    class NumberProperty(CardPropertyABC):
        AcceptableValues = range(1, 101)

    class PokemonCard(CardABC):
        CardProperties = [
            NameProperty('name'),
            NumberProperty('hp'),
            NumberProperty('power'),
        ]

    PokemonCard(
        name='charizard',
        hp=100,
        power=35,
    )

    # mismatched argument
    with pytest.raises(ValueError):
        PokemonCard(
            name=4,
            hp=100,
            power=35,
        )

    # missing values
    with pytest.raises(ValueError):
        PokemonCard(name='Pikachu')





"""Generic library for playing cards
"""

from abc import ABCMeta
from typing import Optional


class CardPropertyABC(metaclass=ABCMeta):
    """Abstract Base Class for all card *property* classes.

    Examples of properties include suit, color, number.
    """
    
    # name of this property e.g. suit
    # PropertyName = None

    # acceptable values for this property e.g. diamond/heart/club/spades
    AcceptableValues = None

    # whether this property must be populated on any given card
    Required = True

    def __init__(self, name):
        self.name = name

    def validate(self, value):
        if self.Required and value is None:
            raise ValueError('Property {} is required'.format(self.name))

        if self.AcceptableValues:
            failed_value_validation = False
            if callable(self.AcceptableValues):
                failed_value_validation = not self.AcceptableValues(value)
            else:
                failed_value_validation = value not in self.AcceptableValues
            if failed_value_validation:
                raise ValueError('"{}" is not an acceptable value for {}'.format(value, self.name))


class CardABC(metaclass=ABCMeta):
    """Abstract Base Class for all card classes."""

    # types of properties on this type of card
    # property_name:  property_value
    CardProperties = []

    @classmethod
    def required_props(cls):
        return [prop.key() for prop in self.CardProperties if prop.required]

    def __init__(self, **kwargs) -> None:

        self._data = {}
        self._properties = {
            prop.name: prop
            for prop in self.CardProperties
        }

        # validate properties passed-in to ctor
        for prop_name, prop in self._properties.items():
            prop_value = kwargs.get(prop_name)
            prop.validate(prop_value)
            self._data[prop_name] = prop_value

    def get(self, prop_name, error_if_missing=True) -> Optional[str]:
        if prop_name not in self._data:
            if error_if_missing:
                raise KeyError('Prop {} not in this card'.format(prop_name))
            else:
                return
        return self._data[prop_name]




from .abstract import (
    CardABC,
    CardPropertyABC,
)


class HanabiColorProperty(CardPropertyABC):
    WHITE = 'white'
    YELLOW = 'yellow'
    GREEN = 'green'
    BLUE = 'blue'
    RED = 'red'

    AcceptableValues = {WHITE, YELLOW, GREEN, BLUE, RED}


class HanabiNumberProperty(CardPropertyABC):
    AcceptableValues = set(range(1, 6))


class HanabiCard(CardABC):
    CardProperties = [
        HanabiColorProperty('color'),
        HanabiNumberProperty('number'),
    ]

    def __init__(self, color, number):
        super().__init__(color=color, number=number)



import pytest

from .hanabi_cards import HanabiCard, HanabiColorProperty


def test_making_hanabi_cards():
    with pytest.raises(TypeError):
        HanabiCard()

    with pytest.raises(ValueError):
        HanabiCard('purple', 1)

    with pytest.raises(ValueError):
        HanabiCard(HanabiColorProperty.YELLOW, 6)

    HanabiCard(HanabiColorProperty.WHITE, 5)
    HanabiCard(HanabiColorProperty.BLUE, 1)




from .abstract import (
    CardABC,
    CardPropertyABC,
)


_VALID_NUMBERS = set(range(2, 11))
_VALID_HIGH_CARDS = {'J', 'Q', 'K', 'A'}
VALID_FACE_VALUES = _VALID_NUMBERS | _VALID_HIGH_CARDS


class PlayingCardNumberProperty(CardPropertyABC):
    AcceptableValues = VALID_FACE_VALUES


class PlayingCardSuitProperty(CardPropertyABC):
    CLUBS = 'clubs'
    DIAMONDS = 'diamonds'
    HEARTS = 'hearts'
    SPADES = 'spades'

    AcceptableValues = {CLUBS, DIAMONDS, HEARTS, SPADES}


class PlayingCard(CardABC):
    CardProperties = [
        PlayingCardNumberProperty('number'),
        PlayingCardSuitProperty('suit'),
    ]
